// File: qsm_linesize.h

// And for all QB64 fellow members who need maximum speed in their sorting, here comes the usage of the "header" written in C, ready to go in QB64:

// Testfile:
// https://dumps.wikimedia.org/enwiki/20211201/enwiki-20211201-pages-articles.xml.bz2 (17.9 GB)

//Declare CustomType Library "qsm_linesize"
//    Sub Quicksort_QB64_v7_linesize (ByVal QWORDSoff As _Offset, ByVal QWORDSlen As _Offset, Byval Left As _Integer64, Byval Right As _Integer64)
//End Declare
//Quicksort_QB64_v7_linesize MhandleOFF.OFFSET, MhandleLEN.OFFSET, 0, ElementsMinusOne
 
/*
Declare CustomType Library
    Sub memcpy (ByVal dest As _Offset, Byval source As _Offset, Byval bytes As Long)
End Declare

a$ = "1234567890"
b$ = "ABCDEFGHIJ"

memcpy _Offset(a$) + 5, _Offset(b$) + 5, 5
Print a$
*/

#include <stdint.h> // Needed for uint64_t

// Quicksort 'Magnetica' r.4, unbreakable (never overflowing the stack) when '#define FinishWithInsertionSort' is uncommented [
// Roughly speaking:
// - with (median of 1) nastiest quadratic O((N/1)^2) we have 2(N/1) (due to Left+Right=2, a pair) stack usage;
// - with (median of 3) less nastier quadratic O((N/2)^2) we have 2(N/2) (due to Left+Right=2, a pair) stack usage;
// - with (median of 5) even less nastier quadratic O((N/3)^2) we have 2(N/3) (due to Left+Right=2, a pair) stack usage;
// - with (median of 7) even more less nastier quadratic O((N/4)^2) we have 2(N/4) (due to Left+Right=2, a pair) stack usage;
// Value below allows 2(N/4)= 2((Right-Left+1)/4) = (99999-1) i.e. N=2(99999-1) or ~200,000 elements
#define StackEntries 99999
// If below define is commented then we have "middle Pivot standalone Magnetica", otherwise "median of 3/7 Pivot InsertionSort Magnetica".
#define FinishWithInsertionSort
// Uncomment either 4 (fastest so far) or 3 (for reference only):
#define revision4
//#define revision3
// Thus we have 4 variants:
// Variant #1: "middle Pivot standalone Magnetica r.4"
// Variant #2: "middle Pivot standalone Magnetica r.3"
// Variant #3: "median of 3/7 Pivot InsertionSort Magnetica r.4"
// Variant #4: "median of 3/7 Pivot InsertionSort Magnetica r.3"
	//void swapUnconditional(uint64_t *a, uint64_t *b) { uint64_t t = *a; *a = *b; *b = t; }
// Many thanks go to all the coders at: https://stackoverflow.com/questions/2786899/fastest-sort-of-fixed-length-6-int-array
#define minSO(x, y) (y ^ ((x ^ y) & -(x < y)))
#define maxSO(x, y) (x ^ ((x ^ y) & -(x < y)))
//    _____                                __  .__               
//   /     \ _____     ____   ____   _____/  |_|__| ____ _____   
//  /  \ /  \\__  \   / ___\ /    \_/ __ \   __\  |/ ___\\__  \  
// /    Y    \/ __ \_/ /_/  >   |  \  ___/|  | |  \  \___ / __ \_
// \____|__  (____  /\___  /|___|  /\___  >__| |__|\___  >____  /
//         \/     \//_____/      \/     \/             \/     \/ 
// Written by Sanmayce, 2021-Dec-02
// The indexes are signed, but the elements are unsigned, targeting 64bit pointers mostly.
void Quicksort_QB64_v7_linesize(uint64_t QWORDSoff[], uint64_t QWORDS[], int64_t Left, int64_t Right) {
    #ifdef FinishWithInsertionSort
    int InsertionsortTHRESHOLD = 17;
    #else
    int InsertionsortTHRESHOLD = 0;
    #endif
    int64_t Indx, Jndx, PL, PR;
    int64_t Stack[StackEntries];
    int64_t StackPtr = 0;
    uint64_t Pivot;
    int64_t LeftBackup = Left;
    int64_t RightBackup = Right;
    register uint64_t x0,x1,x2,x3,x4,x5,x6;
    register uint64_t x0off,x1off,x2off,x3off,x4off,x5off,x6off;
    int o0,o1,o2,o3,o4,o5,o6;
    int64_t TheMiddleOfMiddle;
    int Stefan_Edelkamp_Armin_Weiss1, Stefan_Edelkamp_Armin_Weiss2, Stefan_Edelkamp_Armin_Weiss3;
    uint64_t tmp;
    int GearBox = 0; // 0 is Median of 3; 1 is Median of 7; Shifting to higher gear if BiggerPartition:SmallerPartition ratio is > 64:1
    StackPtr++; Stack[StackPtr] = Left;
    StackPtr++; Stack[StackPtr] = Right;
    do {
        Right = Stack[StackPtr];
        Left = Stack[StackPtr - 1];
        StackPtr = StackPtr - 2;
        //do {
        for(;(Left + InsertionsortTHRESHOLD < Right);) { // For instance, 1 + 17 < 19 i.e. (19-1)/4+6 < 19
            Jndx = Right;
            PL = Left;
            PR = Left;
    #ifdef FinishWithInsertionSort
            TheMiddleOfMiddle = Left + ((Right-Left)>>2); // (4Left + Right - Left)/4; Caution: [TheMiddleOfMiddle+6] should be <= [Right] i.e. 6+(4Left + Right - Left)/4 <= Right or 4*6+(4Left + Right - Left) <= 4Right or 4*6 <= 3Right-3Left or (4*6)/3 <= Right-Left i.e. 8 <= Right-Left as a minimum condition to enter the 'for'.
            x0 = QWORDS[TheMiddleOfMiddle+0];
            x1 = QWORDS[TheMiddleOfMiddle+1];
            x2 = QWORDS[TheMiddleOfMiddle+2];
            x0off = QWORDSoff[TheMiddleOfMiddle+0];
            x1off = QWORDSoff[TheMiddleOfMiddle+1];
            x2off = QWORDSoff[TheMiddleOfMiddle+2];
if (GearBox == 0) { // Median of 5 proves more effective than 7, however 7 betters better the nastiest N^2
            o0 = (x0>x1)+(x0>x2);
            o1 = (x1>=x0)+(x1>x2);
            o2 = (0+1+2)-(o0+o1);
            QWORDS[TheMiddleOfMiddle+o0]=x0; QWORDS[TheMiddleOfMiddle+o1]=x1; QWORDS[TheMiddleOfMiddle+o2]=x2;
            QWORDSoff[TheMiddleOfMiddle+o0]=x0off; QWORDSoff[TheMiddleOfMiddle+o1]=x1off; QWORDSoff[TheMiddleOfMiddle+o2]=x2off;
            swapUnconditional (&QWORDS[TheMiddleOfMiddle+1], &QWORDS[PR]); //2nd
            swapUnconditional (&QWORDSoff[TheMiddleOfMiddle+1], &QWORDSoff[PR]); //2nd
} else {
            x3 = QWORDS[TheMiddleOfMiddle+3];
            x4 = QWORDS[TheMiddleOfMiddle+4];
            x5 = QWORDS[TheMiddleOfMiddle+5];
            x6 = QWORDS[TheMiddleOfMiddle+6];
            x3off = QWORDSoff[TheMiddleOfMiddle+3];
            x4off = QWORDSoff[TheMiddleOfMiddle+4];
            x5off = QWORDSoff[TheMiddleOfMiddle+5];
            x6off = QWORDSoff[TheMiddleOfMiddle+6];
            o0 = (x0>x1)+(x0>x2)+(x0>x3)+(x0>x4)+(x0>x5)+(x0>x6);
            o1 = (x1>=x0)+(x1>x2)+(x1>x3)+(x1>x4)+(x1>x5)+(x1>x6);
            o2 = (x2>=x0)+(x2>=x1)+(x2>x3)+(x2>x4)+(x2>x5)+(x2>x6);
            o3 = (x3>=x0)+(x3>=x1)+(x3>=x2)+(x3>x4)+(x3>x5)+(x3>x6);
            o4 = (x4>=x0)+(x4>=x1)+(x4>=x2)+(x4>=x3)+(x4>x5)+(x4>x6);
            o5 = (x5>=x0)+(x5>=x1)+(x5>=x2)+(x5>=x3)+(x5>=x4)+(x5>x6);
            o6 = (0+1+2+3+4+5+6)-(o0+o1+o2+o3+o4+o5);
            QWORDS[TheMiddleOfMiddle+o0]=x0; QWORDS[TheMiddleOfMiddle+o1]=x1; QWORDS[TheMiddleOfMiddle+o2]=x2; QWORDS[TheMiddleOfMiddle+o3]=x3; QWORDS[TheMiddleOfMiddle+o4]=x4; QWORDS[TheMiddleOfMiddle+o5]=x5; QWORDS[TheMiddleOfMiddle+o6]=x6;
            QWORDSoff[TheMiddleOfMiddle+o0]=x0off; QWORDSoff[TheMiddleOfMiddle+o1]=x1off; QWORDSoff[TheMiddleOfMiddle+o2]=x2off; QWORDSoff[TheMiddleOfMiddle+o3]=x3off; QWORDSoff[TheMiddleOfMiddle+o4]=x4off; QWORDSoff[TheMiddleOfMiddle+o5]=x5off; QWORDSoff[TheMiddleOfMiddle+o6]=x6off;
            swapUnconditional (&QWORDS[TheMiddleOfMiddle+3], &QWORDS[PR]); //4th
            swapUnconditional (&QWORDSoff[TheMiddleOfMiddle+3], &QWORDSoff[PR]); //4th
}
            Pivot = QWORDS[PR];
    #else
            swapUnconditional (&QWORDS[(Left + Right)>>1], &QWORDS[PR]);
            swapUnconditional (&QWORDSoff[(Left + Right)>>1], &QWORDSoff[PR]);
            Pivot = QWORDS[PR];
    #endif
    #ifdef revision4
            for (;PR < Jndx;) {
                PR = PR + 1;
                if (Pivot > QWORDS[PR]) {
                    swapUnconditional (&QWORDS[PL], &QWORDS[PR]);
                    swapUnconditional (&QWORDSoff[PL], &QWORDSoff[PR]);
                    PL = PL + 1;
                //} else if (Pivot == QWORDS[PR]) {
                } else if (Pivot < QWORDS[PR]) {
                    for (;Pivot < QWORDS[Jndx];) {
                        Jndx = Jndx - 1;
                    }
                    if (PR < Jndx) { swapUnconditional (&QWORDS[PR], &QWORDS[Jndx]); swapUnconditional (&QWORDSoff[PR], &QWORDSoff[Jndx]); }
                    Jndx = Jndx - 1;
                    PR = PR - 1;
                }
            }
    #endif
/*
; mark_description "Intel(R) C++ Compiler XE for applications running on Intel(R) 64, Version 15.0.0.108 Build 20140726";
; ae-64+2=76 bytes i.e. (94-76=18 less), 5/1 conditional/unconditional jumps i.e. 2-1=1 less unconditional jump than r.1
; 'Magnetica' partitioning, mainloop rev.2[
.B4.5::                         
  00064 48 ff c5         inc rbp                                
  00067 48 8b 14 e9      mov rdx, QWORD PTR [rcx+rbp*8]         
  0006b 4c 3b ea         cmp r13, rdx                           
  0006e 77 2c            ja .B4.14 
.B4.6::                         
  00070 73 39            jae .B4.15 
.B4.7::                         
  00072 4e 8b 3c d9      mov r15, QWORD PTR [rcx+r11*8]         
  00076 4d 3b ef         cmp r13, r15                           
  00079 73 0c            jae .B4.11 
.B4.9::                         
  0007b 49 ff cb         dec r11                                
  0007e 4e 8b 3c d9      mov r15, QWORD PTR [rcx+r11*8]         
  00082 4d 3b ef         cmp r13, r15                           
  00085 72 f4            jb .B4.9 
.B4.11::                        
  00087 49 3b eb         cmp rbp, r11                           
  0008a 7d 08            jge .B4.13 
.B4.12::                        
  0008c 4c 89 3c e9      mov QWORD PTR [rcx+rbp*8], r15         
  00090 4a 89 14 d9      mov QWORD PTR [rcx+r11*8], rdx         
.B4.13::                        
  00094 49 ff cb         dec r11                                
  00097 48 ff cd         dec rbp                                
  0009a eb 0f            jmp .B4.15 
.B4.14::                        
  0009c 4e 8b 3c f1      mov r15, QWORD PTR [rcx+r14*8]         
  000a0 4a 89 14 f1      mov QWORD PTR [rcx+r14*8], rdx         
  000a4 49 ff c6         inc r14                                
  000a7 4c 89 3c e9      mov QWORD PTR [rcx+rbp*8], r15         
.B4.15::                        
  000ab 49 3b eb         cmp rbp, r11                           
  000ae 7c b4            jl .B4.5 
; 'Magnetica' partitioning, mainloop rev.2]
*/
    #ifdef revision3
            for (;PR < Jndx;) {
                // Many thanks go to Orson Peters for sharing his Pattern-defeating quicksort (pdqsort) at GitHub.
                // This is due to a new technique described in "BlockQuicksort: How Branch Mispredictions don't affect Quicksort" by Stefan Edelkamp and Armin Weiss.
                Stefan_Edelkamp_Armin_Weiss1 = (Pivot > QWORDS[PR + 1]);
                Stefan_Edelkamp_Armin_Weiss2 = (Pivot == QWORDS[PR + 1]);
                Stefan_Edelkamp_Armin_Weiss3 = (Pivot < QWORDS[PR + 1]);
                tmp = minSO(QWORDS[PL], QWORDS[PR + 1]); QWORDS[PR + 1] = maxSO(QWORDS[PL], QWORDS[PR + 1]); QWORDS[PL] = tmp; // Since Pivot is equal to QWORDS[PL]
                PL = PL + Stefan_Edelkamp_Armin_Weiss1;
                PR = PR + Stefan_Edelkamp_Armin_Weiss1;
                PR = PR + Stefan_Edelkamp_Armin_Weiss2;
                if (Stefan_Edelkamp_Armin_Weiss3) {
                    for (;Pivot < QWORDS[Jndx];) {
                        Jndx = Jndx - 1;
                    }
                    if (PR + 1 < Jndx) swapUnconditional (&QWORDS[PR + 1], &QWORDS[Jndx]);
                    Jndx = Jndx - 1;
                }
            }
    #endif
/*
; mark_description "Intel(R) C++ Compiler XE for applications running on Intel(R) 64, Version 15.0.0.108 Build 20140726";
; f9-73+6=140 bytes, 4/0 conditional/unconditional jumps
; 'Magnetica' partitioning, mainloop rev.3[
.B7.5::                         
  00073 4a 8b 5c f1 08   mov rbx, QWORD PTR [8+rcx+r14*8]       
  00078 33 d2            xor edx, edx                           
  0007a 4c 3b eb         cmp r13, rbx                           
  0007d 4a 8b 2c c1      mov rbp, QWORD PTR [rcx+r8*8]          
  00081 49 0f 47 d4      cmova rdx, r12                         
  00085 45 33 ff         xor r15d, r15d                         
  00088 48 3b eb         cmp rbp, rbx                           
  0008b 48 89 ee         mov rsi, rbp                           
  0008e 41 0f 92 c7      setb r15b                              
  00092 48 33 f3         xor rsi, rbx                           
  00095 41 f7 df         neg r15d                               
  00098 4d 63 ff         movsxd r15, r15d                       
  0009b 49 23 f7         and rsi, r15                           
  0009e 48 33 ee         xor rbp, rsi                           
  000a1 4c 3b eb         cmp r13, rbx                           
  000a4 4a 89 6c f1 08   mov QWORD PTR [8+rcx+r14*8], rbp       
  000a9 bd 00 00 00 00   mov ebp, 0                             
  000ae 49 0f 44 ec      cmove rbp, r12                         
  000b2 48 33 f3         xor rsi, rbx                           
  000b5 4a 89 34 c1      mov QWORD PTR [rcx+r8*8], rsi          
  000b9 4c 03 c2         add r8, rdx                            
  000bc 48 03 d5         add rdx, rbp                           
  000bf 4c 03 f2         add r14, rdx                           
  000c2 4c 3b eb         cmp r13, rbx                           
  000c5 73 2f            jae .B7.13 
.B7.6::                         
  000c7 4a 8b 14 c9      mov rdx, QWORD PTR [rcx+r9*8]          
  000cb 4c 3b ea         cmp r13, rdx                           
  000ce 73 0c            jae .B7.10 
.B7.8::                         
  000d0 49 ff c9         dec r9                                 
  000d3 4a 8b 14 c9      mov rdx, QWORD PTR [rcx+r9*8]          
  000d7 4c 3b ea         cmp r13, rdx                           
  000da 72 f4            jb .B7.8 
.B7.10::                        
  000dc 49 8d 5e 01      lea rbx, QWORD PTR [1+r14]             
  000e0 4c 3b cb         cmp r9, rbx                            
  000e3 7e 0e            jle .B7.12 
.B7.11::                        
  000e5 4a 8b 5c f1 08   mov rbx, QWORD PTR [8+rcx+r14*8]       
  000ea 4a 89 54 f1 08   mov QWORD PTR [8+rcx+r14*8], rdx       
  000ef 4a 89 1c c9      mov QWORD PTR [rcx+r9*8], rbx          
.B7.12::                        ; Preds .B7.10 .B7.11
  000f3 49 ff c9         dec r9                                 
.B7.13::                        
  000f6 4d 3b f1         cmp r14, r9                            
  000f9 0f 8c 74 ff ff 
        ff               jl .B7.5 
; 'Magnetica' partitioning, mainloop rev.3]
*/
            Jndx = PL - 1;
            Indx = PR + 1;
/*
            // 'Magnetica' partitioning [
            Jndx = Right;
            PL = Left;
            PR = Left;
            swap (&QWORDS[(Left + Right)>>1], &QWORDS[PR]);
            Pivot = QWORDS[PR];
            for (;PR < Jndx;) {
                if (Pivot > QWORDS[PR + 1]) {
                    swap (&QWORDS[PL], &QWORDS[PR + 1]);
                    PL = PL + 1;
                    PR = PR + 1;
                } else if (Pivot == QWORDS[PR + 1]) {
                    PR = PR + 1;
                } else {
                    for (;Pivot < QWORDS[Jndx];) {
                        Jndx = Jndx - 1;
                    }
                    if (PR + 1 < Jndx) swap (&QWORDS[PR + 1], &QWORDS[Jndx]);
                    Jndx = Jndx - 1;
                }
            }
            Jndx = PL - 1;
            Indx = PR + 1;
            // 'Magnetica' partitioning ]
*/
/*
; mark_description "Intel(R) C++ Compiler XE for applications running on Intel(R) 64, Version 15.0.0.108 Build 20140726";
; c6-6a+2=94 bytes, 5/2 conditional/unconditional jumps
; 'Magnetica' partitioning, mainloop [
.B4.5::                         
  0006a 4e 3b 5c e9 08   cmp r11, QWORD PTR [8+rcx+r13*8]       
  0006f 77 37            ja .B4.15 
.B4.6::                         
  00071 75 08            jne .B4.8 
.B4.7::                         
  00073 49 89 d5         mov r13, rdx                           
  00076 48 ff c2         inc rdx                                
  00079 eb 48            jmp .B4.16 
.B4.8::                         
  0007b 4e 8b 34 c1      mov r14, QWORD PTR [rcx+r8*8]          
  0007f 4d 3b de         cmp r11, r14                           
  00082 73 0c            jae .B4.12 
.B4.10::                        
  00084 49 ff c8         dec r8                                 
  00087 4e 8b 34 c1      mov r14, QWORD PTR [rcx+r8*8]          
  0008b 4d 3b de         cmp r11, r14                           
  0008e 72 f4            jb .B4.10 
.B4.12::                        
  00090 4c 3b c2         cmp r8, rdx                            
  00093 7e 0e            jle .B4.14 
.B4.13::                        
  00095 4e 8b 7c e9 08   mov r15, QWORD PTR [8+rcx+r13*8]       
  0009a 4e 89 74 e9 08   mov QWORD PTR [8+rcx+r13*8], r14       
  0009f 4e 89 3c c1      mov QWORD PTR [rcx+r8*8], r15          
.B4.14::                        
  000a3 49 ff c8         dec r8                                 
  000a6 eb 1b            jmp .B4.16 
.B4.15::                        
  000a8 4e 8b 74 e9 08   mov r14, QWORD PTR [8+rcx+r13*8]       
  000ad 4e 8b 3c e1      mov r15, QWORD PTR [rcx+r12*8]         
  000b1 4e 89 34 e1      mov QWORD PTR [rcx+r12*8], r14         
  000b5 49 ff c4         inc r12                                
  000b8 4e 89 7c e9 08   mov QWORD PTR [8+rcx+r13*8], r15       
  000bd 49 89 d5         mov r13, rdx                           
  000c0 48 ff c2         inc rdx                                
.B4.16::                        
  000c3 4d 3b e8         cmp r13, r8                            
  000c6 7c a2            jl .B4.5 
; 'Magnetica' partitioning, mainloop ]
*/
    #ifdef FinishWithInsertionSort
            GearBox = ( maxSO((Right-Indx), (Jndx-Left)) > minSO((Right-Indx), (Jndx-Left))<<6 ); // same as MAX/MIN > 64, i.e. Gear=1 if we need Median of 7
    #endif
            if (Indx + InsertionsortTHRESHOLD < Right) { // still refusing to go (always) with the smaller partition first...
                StackPtr = StackPtr + 2;
                Stack[StackPtr - 1] = Indx;
                Stack[StackPtr] = Right;
    #ifdef FinishWithInsertionSort
                // Stack is full, bail out and finalize with Insertionsort [
                StackPtr = StackPtr*(StackPtr + 2 <= StackEntries-1);  // StackPtr should be enforced FALSE; also, ensure the next 'push' above is sanctioned - the max index of Stack[] being 'StackEntries-1'
                Right = Right*(StackPtr + 2 <= StackEntries-1); // Left + InsertionsortTHRESHOLD < Right should be enforced FALSE:
                // Stack is full, bail out and finalize with Insertionsort ]
    #endif
            }
            Right = Jndx;
        } //while (Left + InsertionsortTHRESHOLD < Right);
    } while ( (StackPtr != 0) );
    #ifdef FinishWithInsertionSort
    for (Indx=LeftBackup+1; Indx <= RightBackup; Indx++) {
        Jndx = Indx;
        for (;Jndx >= 1;) {
            if (QWORDS[Jndx-1] > QWORDS[Jndx]) { swapUnconditional (&QWORDS[Jndx-1], &QWORDS[Jndx]); swapUnconditional (&QWORDSoff[Jndx-1], &QWORDSoff[Jndx]);}
            else break;
            Jndx = Jndx - 1;
        }
    }
    #endif
}
// My threads with Magnetica's source and binaries (Linux and Windows):
// https://www.overclock.net/threads/benchmark-quicksort-says-sorting-2-billion-qwords.1794855/
// https://www.qb64.org/forum/index.php?topic=3518.msg137645#msg137645
// https://www.linuxquestions.org/questions/programming-9/qsort-vs-%27magnetica%27-quicksort-4175703333/#post6299782
// Quicksort 'Magnetica' r.4, unbreakable (never overflowing the stack) when '#define FinishWithInsertionSort' is uncommented ]

// 113 lines of scalar C:
/*
00,001 #define StackEntries 99999
00,002 #define FinishWithInsertionSort
00,003 #define revision4
00,004 void swapUnconditional(uint64_t *a, uint64_t *b) { uint64_t t = *a; *a = *b; *b = t; }
00,005 #define minSO(x, y) (y ^ ((x ^ y) & -(x < y)))
00,006 #define maxSO(x, y) (x ^ ((x ^ y) & -(x < y)))
00,007 //    _____                                __  .__               
00,008 //   /     \ _____     ____   ____   _____/  |_|__| ____ _____   
00,009 //  /  \ /  \\__  \   / ___\ /    \_/ __ \   __\  |/ ___\\__  \  
00,010 // /    Y    \/ __ \_/ /_/  >   |  \  ___/|  | |  \  \___ / __ \_
00,011 // \____|__  (____  /\___  /|___|  /\___  >__| |__|\___  >____  /
00,012 //         \/     \//_____/      \/     \/             \/     \/ 
00,013 // Written by Sanmayce, 2021-Dec-02
00,014 void Quicksort_QB64_v7(uint64_t QWORDS[], int64_t Left, int64_t Right) {
00,015     #ifdef FinishWithInsertionSort
00,016     int InsertionsortTHRESHOLD = 17;
00,017     #else
00,018     int InsertionsortTHRESHOLD = 0;
00,019     #endif
00,020     int64_t Indx, Jndx, PL, PR;
00,021     int64_t Stack[StackEntries];
00,022     int64_t StackPtr = 0;
00,023     uint64_t Pivot;
00,024     int64_t LeftBackup = Left;
00,025     int64_t RightBackup = Right;
00,026     register uint64_t x0,x1,x2,x3,x4,x5,x6;
00,027     int o0,o1,o2,o3,o4,o5,o6;
00,028     int64_t TheMiddleOfMiddle;
00,029     int GearBox = 0; // 0 is Median of 3; 1 is Median of 7; Shifting to higher gear if BiggerPartition:SmallerPartition ratio is > 64:1
00,030     StackPtr++; Stack[StackPtr] = Left;
00,031     StackPtr++; Stack[StackPtr] = Right;
00,032     do {
00,033         Right = Stack[StackPtr];
00,034         Left = Stack[StackPtr - 1];
00,035         StackPtr = StackPtr - 2;
00,036         for(;(Left + InsertionsortTHRESHOLD < Right);) { // For instance, 1 + 17 < 19 i.e. (19-1)/4+6 < 19
00,037             Jndx = Right;
00,038             PL = Left;
00,039             PR = Left;
00,040     #ifdef FinishWithInsertionSort
00,041             TheMiddleOfMiddle = Left + ((Right-Left)>>2); // (4Left + Right - Left)/4; Caution: [TheMiddleOfMiddle+6] should be <= [Right] i.e. 6+(4Left + Right - Left)/4 <= Right or 4*6+(4Left + Right - Left) <= 4Right or 4*6 <= 3Right-3Left or (4*6)/3 <= Right-Left i.e. 8 <= Right-Left as a minimum condition to enter the 'for'.
00,042             x0 = QWORDS[TheMiddleOfMiddle+0];
00,043             x1 = QWORDS[TheMiddleOfMiddle+1];
00,044             x2 = QWORDS[TheMiddleOfMiddle+2];
00,045 if (GearBox == 0) { // Median of 5 proves more effective than 7, however 7 betters better the nastiest N^2
00,046             o0 = (x0>x1)+(x0>x2);
00,047             o1 = (x1>=x0)+(x1>x2);
00,048             o2 = (0+1+2)-(o0+o1);
00,049             QWORDS[TheMiddleOfMiddle+o0]=x0; QWORDS[TheMiddleOfMiddle+o1]=x1; QWORDS[TheMiddleOfMiddle+o2]=x2;
00,050             swapUnconditional (&QWORDS[TheMiddleOfMiddle+1], &QWORDS[PR]); //2nd
00,051 } else {
00,052             x3 = QWORDS[TheMiddleOfMiddle+3];
00,053             x4 = QWORDS[TheMiddleOfMiddle+4];
00,054             x5 = QWORDS[TheMiddleOfMiddle+5];
00,055             x6 = QWORDS[TheMiddleOfMiddle+6];
00,056             o0 = (x0>x1)+(x0>x2)+(x0>x3)+(x0>x4)+(x0>x5)+(x0>x6);
00,057             o1 = (x1>=x0)+(x1>x2)+(x1>x3)+(x1>x4)+(x1>x5)+(x1>x6);
00,058             o2 = (x2>=x0)+(x2>=x1)+(x2>x3)+(x2>x4)+(x2>x5)+(x2>x6);
00,059             o3 = (x3>=x0)+(x3>=x1)+(x3>=x2)+(x3>x4)+(x3>x5)+(x3>x6);
00,060             o4 = (x4>=x0)+(x4>=x1)+(x4>=x2)+(x4>=x3)+(x4>x5)+(x4>x6);
00,061             o5 = (x5>=x0)+(x5>=x1)+(x5>=x2)+(x5>=x3)+(x5>=x4)+(x5>x6);
00,062             o6 = (0+1+2+3+4+5+6)-(o0+o1+o2+o3+o4+o5);
00,063             QWORDS[TheMiddleOfMiddle+o0]=x0; QWORDS[TheMiddleOfMiddle+o1]=x1; QWORDS[TheMiddleOfMiddle+o2]=x2; QWORDS[TheMiddleOfMiddle+o3]=x3; QWORDS[TheMiddleOfMiddle+o4]=x4; QWORDS[TheMiddleOfMiddle+o5]=x5; QWORDS[TheMiddleOfMiddle+o6]=x6;
00,064             swapUnconditional (&QWORDS[TheMiddleOfMiddle+3], &QWORDS[PR]); //4th
00,065 }
00,066             Pivot = QWORDS[PR];
00,067     #else
00,068             swapUnconditional (&QWORDS[(Left + Right)>>1], &QWORDS[PR]);
00,069             Pivot = QWORDS[PR];
00,070     #endif
00,071     #ifdef revision4
00,072             for (;PR < Jndx;) {
00,073                 PR = PR + 1;
00,074                 if (Pivot > QWORDS[PR]) {
00,075                     swapUnconditional (&QWORDS[PL], &QWORDS[PR]);
00,076                     PL = PL + 1;
00,077                 } else if (Pivot < QWORDS[PR]) {
00,078                     for (;Pivot < QWORDS[Jndx];) {
00,079                         Jndx = Jndx - 1;
00,080                     }
00,081                     if (PR < Jndx) swapUnconditional (&QWORDS[PR], &QWORDS[Jndx]);
00,082                     Jndx = Jndx - 1;
00,083                     PR = PR - 1;
00,084                 }
00,085             }
00,086     #endif
00,087             Jndx = PL - 1;
00,088             Indx = PR + 1;
00,089     #ifdef FinishWithInsertionSort
00,090             GearBox = ( maxSO((Right-Indx), (Jndx-Left)) > minSO((Right-Indx), (Jndx-Left))<<6 ); // same as MAX/MIN > 64, i.e. Gear=1 if we need Median of 7
00,091     #endif
00,092             if (Indx + InsertionsortTHRESHOLD < Right) { // still refusing to go (always) with the smaller partition first...
00,093                 StackPtr = StackPtr + 2;
00,094                 Stack[StackPtr - 1] = Indx;
00,095                 Stack[StackPtr] = Right;
00,096     #ifdef FinishWithInsertionSort
00,097                 StackPtr = StackPtr*(StackPtr + 2 <= StackEntries-1);  // StackPtr should be enforced FALSE; also, ensure the next 'push' above is sanctioned - the max index of Stack[] being 'StackEntries-1'
00,098                 Right = Right*(StackPtr + 2 <= StackEntries-1); // Left + InsertionsortTHRESHOLD < Right should be enforced FALSE:
00,099     #endif
00,100             }
00,101             Right = Jndx;
00,102         } //while (Left + InsertionsortTHRESHOLD < Right);
00,103     } while ( (StackPtr != 0) );
00,104     #ifdef FinishWithInsertionSort
00,105     for (Indx=LeftBackup+1; Indx <= RightBackup; Indx++) {
00,106         Jndx = Indx;
00,107         for (;Jndx >= 1;) {
00,108             if (QWORDS[Jndx-1] > QWORDS[Jndx]) swapUnconditional (&QWORDS[Jndx-1], &QWORDS[Jndx]); else break;
00,109             Jndx = Jndx - 1;
00,110         }
00,111     }
00,112     #endif
00,113 }
*/

// 380 lines scalar Assembly:
/*
00,001 ; mark_description "Intel(R) C++ Compiler XE for applications running on Intel(R) 64, Version 15.0.0.108 Build 20140726";
00,002 ; mark_description "-S -Os";
00,003 
00,004 Quicksort_QB64_v7	PROC 
00,005 .B4.1::                         
00,006         push      rbx                                           
00,007         push      rsi                                           
00,008         push      rdi                                           
00,009         push      r12                                           
00,010         push      r13                                           
00,011         push      r14                                           
00,012         push      r15                                           
00,013         push      rbp                                           
00,014         mov       eax, 800152                                   
00,015         call      __chkstk                                      
00,016         sub       rsp, 800152                                   
00,017         mov       r13, rcx                                      
00,018         mov       QWORD PTR [800064+rsp], r8                    
00,019         xor       r14d, r14d                                    
00,020         mov       QWORD PTR [40+rsp], rdx                       
00,021         mov       QWORD PTR [56+rsp], rdx                       
00,022         mov       QWORD PTR [32+rsp], 2                         
00,023         mov       QWORD PTR [64+rsp], r8                        
00,024 .B4.2::                         
00,025         mov       rax, QWORD PTR [32+rsp]                       
00,026         mov       rdx, QWORD PTR [40+rsp+rax*8]                 
00,027         mov       r12, QWORD PTR [48+rsp+rax*8]                 
00,028         add       rax, -2                                       
00,029         mov       QWORD PTR [800040+rsp], rdx                   
00,030         mov       QWORD PTR [32+rsp], rax                       
00,031         lea       rcx, QWORD PTR [17+rdx]                       
00,032         mov       QWORD PTR [800072+rsp], rcx                   
00,033         cmp       r12, rcx                                      
00,034         jle       .B4.27        
00,035 .B4.3::                         
00,036         mov       QWORD PTR [800056+rsp], rdx                   
00,037         lea       rax, QWORD PTR [r13+rdx*8]                    
00,038         mov       QWORD PTR [800048+rsp], rax                   
00,039 .B4.4::                         
00,040         mov       r8, r12                                       
00,041         xor       eax, eax                                      
00,042         mov       rbp, QWORD PTR [800040+rsp]                   
00,043         sub       r8, rbp                                       
00,044         sar       r8, 2                                         
00,045         xor       edx, edx                                      
00,046         add       r8, rbp                                       
00,047         xor       r15d, r15d                                    
00,048         mov       rbx, QWORD PTR [800056+rsp]                   
00,049         mov       rdi, r12                                      
00,050         mov       QWORD PTR [800056+rsp], rbp                   
00,051         lea       rsi, QWORD PTR [rbp*8]                        
00,052         lea       rcx, QWORD PTR [r13+r8*8]                     
00,053         mov       r10, QWORD PTR [rcx]                          
00,054         mov       r11, QWORD PTR [8+rcx]                        
00,055         cmp       r10, r11                                      
00,056         mov       r9, QWORD PTR [16+rcx]                        
00,057         seta      al                                            
00,058         cmp       r10, r9                                       
00,059         seta      dl                                            
00,060         add       eax, edx                                      
00,061         xor       edx, edx                                      
00,062         cmp       r11, r10                                      
00,063         setae     dl                                            
00,064         cmp       r11, r9                                       
00,065         seta      r15b                                          
00,066         add       edx, r15d                                     
00,067         test      r14d, r14d                                    
00,068         jne       .B4.6         
00,069 .B4.5::                         
00,070         add       rcx, 8                                        
00,071         lea       r14, QWORD PTR [r8+rax]                       
00,072         mov       QWORD PTR [r13+r14*8], r10                    
00,073         lea       r10, QWORD PTR [r8+rdx]                       
00,074         sub       r8, rax                                       
00,075         sub       r8, rdx                                       
00,076         mov       QWORD PTR [r13+r10*8], r11                    
00,077         mov       QWORD PTR [24+r13+r8*8], r9                   
00,078         jmp       .B4.7         
00,079 .B4.6::                         
00,080         mov       QWORD PTR [800104+rsp], rsi                   
00,081         xor       r14d, r14d                                    
00,082         mov       rsi, QWORD PTR [24+rcx]                       
00,083         cmp       r10, rsi                                      
00,084         mov       QWORD PTR [800096+rsp], rdi                   
00,085         mov       rdi, QWORD PTR [32+rcx]                       
00,086         seta      r14b                                          
00,087         xor       r15d, r15d                                    
00,088         cmp       r10, rdi                                      
00,089         mov       QWORD PTR [800088+rsp], r12                   
00,090         push      0                                             
00,091         pop       r12                                           
00,092         seta      r12b                                          
00,093         mov       QWORD PTR [800112+rsp], rbp                   
00,094         add       r14d, r12d                                    
00,095         mov       rbp, QWORD PTR [40+rcx]                       
00,096         xor       r12d, r12d                                    
00,097         cmp       r10, rbp                                      
00,098         mov       QWORD PTR [800120+rsp], rbx                   
00,099         seta      r12b                                          
00,100         mov       rbx, QWORD PTR [48+rcx]                       
00,101         cmp       r10, rbx                                      
00,102         mov       QWORD PTR [800080+rsp], rcx                   
00,103         push      0                                             
00,104         pop       rcx                                           
00,105         seta      cl                                            
00,106         add       eax, r14d                                     
00,107         add       r12d, ecx                                     
00,108         xor       ecx, ecx                                      
00,109         cmp       r11, rsi                                      
00,110         seta      cl                                            
00,111         xor       r14d, r14d                                    
00,112         cmp       r11, rdi                                      
00,113         seta      r15b                                          
00,114         cmp       r11, rbp                                      
00,115         seta      r14b                                          
00,116         add       eax, r12d                                     
00,117         xor       r12d, r12d                                    
00,118         cmp       r11, rbx                                      
00,119         seta      r12b                                          
00,120         add       ecx, r15d                                     
00,121         add       edx, ecx                                      
00,122         xor       ecx, ecx                                      
00,123         cmp       r9, r10                                       
00,124         setae     cl                                            
00,125         xor       r15d, r15d                                    
00,126         add       r14d, r12d                                    
00,127         xor       r12d, r12d                                    
00,128         cmp       r9, r11                                       
00,129         setae     r12b                                          
00,130         add       edx, r14d                                     
00,131         add       ecx, r12d                                     
00,132         xor       r12d, r12d                                    
00,133         cmp       r9, rsi                                       
00,134         seta      r12b                                          
00,135         xor       r14d, r14d                                    
00,136         cmp       r9, rdi                                       
00,137         seta      r15b                                          
00,138         add       r12d, r15d                                    
00,139         xor       r15d, r15d                                    
00,140         add       ecx, r12d                                     
00,141         xor       r12d, r12d                                    
00,142         cmp       r9, rbp                                       
00,143         seta      r12b                                          
00,144         cmp       r9, rbx                                       
00,145         seta      r14b                                          
00,146         cmp       rsi, r10                                      
00,147         setae     r15b                                          
00,148         add       r12d, r14d                                    
00,149         xor       r14d, r14d                                    
00,150         cmp       rsi, r11                                      
00,151         setae     r14b                                          
00,152         add       ecx, r12d                                     
00,153         add       r15d, r14d                                    
00,154         xor       r14d, r14d                                    
00,155         cmp       rsi, r9                                       
00,156         setae     r14b                                          
00,157         xor       r12d, r12d                                    
00,158         cmp       rsi, rdi                                      
00,159         seta      r12b                                          
00,160         add       r14d, r12d                                    
00,161         xor       r12d, r12d                                    
00,162         add       r15d, r14d                                    
00,163         xor       r14d, r14d                                    
00,164         cmp       rsi, rbp                                      
00,165         seta      r14b                                          
00,166         cmp       rsi, rbx                                      
00,167         seta      r12b                                          
00,168         add       r14d, r12d                                    
00,169         xor       r12d, r12d                                    
00,170         add       r15d, r14d                                    
00,171         xor       r14d, r14d                                    
00,172         cmp       rdi, r10                                      
00,173         setae     r14b                                          
00,174         cmp       rdi, r11                                      
00,175         setae     r12b                                          
00,176         add       r14d, r12d                                    
00,177         xor       r12d, r12d                                    
00,178         cmp       rdi, r9                                       
00,179         setae     r12b                                          
00,180         cmp       rdi, rsi                                      
00,181         mov       QWORD PTR [800128+rsp], r15                   
00,182         push      0                                             
00,183         pop       r15                                           
00,184         setae     r15b                                          
00,185         add       r12d, r15d                                    
00,186         xor       r15d, r15d                                    
00,187         add       r14d, r12d                                    
00,188         xor       r12d, r12d                                    
00,189         cmp       rdi, rbp                                      
00,190         seta      r12b                                          
00,191         cmp       rdi, rbx                                      
00,192         seta      r15b                                          
00,193         add       r12d, r15d                                    
00,194         xor       r15d, r15d                                    
00,195         add       r14d, r12d                                    
00,196         xor       r12d, r12d                                    
00,197         cmp       rbp, r10                                      
00,198         setae     r12b                                          
00,199         cmp       rbp, r11                                      
00,200         setae     r15b                                          
00,201         add       r12d, r15d                                    
00,202         xor       r15d, r15d                                    
00,203         cmp       rbp, r9                                       
00,204         setae     r15b                                          
00,205         cmp       rbp, rsi                                      
00,206         mov       QWORD PTR [800136+rsp], r14                   
00,207         push      0                                             
00,208         pop       r14                                           
00,209         setae     r14b                                          
00,210         add       r15d, r14d                                    
00,211         xor       r14d, r14d                                    
00,212         add       r12d, r15d                                    
00,213         xor       r15d, r15d                                    
00,214         cmp       rbp, rdi                                      
00,215         setae     r15b                                          
00,216         cmp       rbp, rbx                                      
00,217         seta      r14b                                          
00,218         add       r15d, r14d                                    
00,219         lea       r14, QWORD PTR [r8+rax]                       
00,220         mov       QWORD PTR [r13+r14*8], r10                    
00,221         lea       r10, QWORD PTR [r8+rdx]                       
00,222         mov       QWORD PTR [r13+r10*8], r11                    
00,223         lea       r11, QWORD PTR [r8+rcx]                       
00,224         mov       r10, QWORD PTR [800128+rsp]                   
00,225         add       r12d, r15d                                    
00,226         mov       QWORD PTR [r13+r11*8], r9                     
00,227         add       rcx, r10                                      
00,228         mov       r11, QWORD PTR [800136+rsp]                   
00,229         lea       r9, QWORD PTR [r8+r10]                        
00,230         mov       QWORD PTR [r13+r9*8], rsi                     
00,231         lea       rsi, QWORD PTR [r8+r11]                       
00,232         add       r11, r12                                      
00,233         mov       QWORD PTR [r13+rsi*8], rdi                    
00,234         add       rcx, r11                                      
00,235         mov       rsi, QWORD PTR [800104+rsp]                   
00,236         lea       rdi, QWORD PTR [r8+r12]                       
00,237         sub       r8, rax                                       
00,238         sub       r8, rdx                                       
00,239         sub       r8, rcx                                       
00,240         mov       QWORD PTR [r13+rdi*8], rbp                    
00,241         mov       rcx, QWORD PTR [800080+rsp]                   
00,242         mov       rbp, QWORD PTR [800112+rsp]                   
00,243         add       rcx, 24                                       
00,244         mov       QWORD PTR [168+r13+r8*8], rbx                 
00,245         mov       rbx, QWORD PTR [800120+rsp]                   
00,246         mov       rdi, QWORD PTR [800096+rsp]                   
00,247         mov       r12, QWORD PTR [800088+rsp]                   
00,248 .B4.7::                         
00,249         mov       rdx, QWORD PTR [800048+rsp]                   
00,250         call      swapUnconditional                             
00,251 .B4.8::                         
00,252         mov       rax, QWORD PTR [800048+rsp]                   
00,253         cmp       r12, QWORD PTR [800040+rsp]                   
00,254         mov       r14, QWORD PTR [rax]                          
00,255         jle       .B4.23        
00,256 .B4.10::                        
00,257         inc       rbp                                           
00,258         lea       rdx, QWORD PTR [r13+rbp*8]                    
00,259         cmp       r14, QWORD PTR [rdx]                          
00,260         jbe       .B4.13        
00,261 .B4.11::                        
00,262         lea       rcx, QWORD PTR [r13+rsi]                      
00,263         call      swapUnconditional                             
00,264 .B4.12::                        
00,265         add       rsi, 8                                        
00,266         inc       rbx                                           
00,267         jmp       .B4.21        
00,268 .B4.13::                        
00,269         jae       .B4.21        
00,270 .B4.14::                        
00,271         lea       rax, QWORD PTR [r13+rdi*8]                    
00,272         cmp       r14, QWORD PTR [rax]                          
00,273         jae       .B4.18        
00,274 .B4.16::                        
00,275         dec       rdi                                           
00,276         lea       rax, QWORD PTR [r13+rdi*8]                    
00,277         cmp       r14, QWORD PTR [rax]                          
00,278         jb        .B4.16        
00,279 .B4.18::                        
00,280         cmp       rbp, rdi                                      
00,281         jge       .B4.20        
00,282 .B4.19::                        
00,283         mov       rcx, rdx                                      
00,284         mov       rdx, rax                                      
00,285         call      swapUnconditional                             
00,286 .B4.20::                        
00,287         dec       rdi                                           
00,288         dec       rbp                                           
00,289 .B4.21::                        
00,290         cmp       rbp, rdi                                      
00,291         jl        .B4.10        
00,292 .B4.23::                        
00,293         dec       rbx                                           
00,294         lea       r9, QWORD PTR [1+rbp]                         
00,295         mov       rdi, r12                                      
00,296         mov       r8, rbx                                       
00,297         sub       r8, QWORD PTR [800040+rsp]                    
00,298         sub       rdi, r9                                       
00,299         xor       edx, edx                                      
00,300         cmp       rdi, r8                                       
00,301         mov       rcx, rdi                                      
00,302         setl      dl                                            
00,303         xor       r14d, r14d                                    
00,304         xor       rcx, r8                                       
00,305         neg       edx                                           
00,306         add       rbp, 18                                       
00,307         movsxd    rdx, edx                                      
00,308         and       rcx, rdx                                      
00,309         xor       r8, rcx                                       
00,310         xor       rdi, rcx                                      
00,311         shl       r8, 6                                         
00,312         cmp       rdi, r8                                       
00,313         setg      r14b                                          
00,314         cmp       r12, rbp                                      
00,315         jle       .B4.25        
00,316 .B4.24::                        
00,317         mov       rax, QWORD PTR [32+rsp]                       
00,318         cmp       rax, 99994                                    
00,319         mov       QWORD PTR [64+rsp+rax*8], r12                 
00,320         push      0                                             
00,321         pop       r12                                           
00,322         setle     r12b                                          
00,323         mov       QWORD PTR [56+rsp+rax*8], r9                  
00,324         add       rax, 2                                        
00,325         imul      rax, r12                                      
00,326         mov       QWORD PTR [32+rsp], rax                       
00,327 .B4.25::                        
00,328         mov       r12, rbx                                      
00,329         cmp       rbx, QWORD PTR [800072+rsp]                   
00,330         jg        .B4.4         
00,331 .B4.27::                        
00,332         cmp       QWORD PTR [32+rsp], 0                         
00,333         jne       .B4.2         
00,334 .B4.28::                        
00,335         mov       rax, QWORD PTR [40+rsp]                       
00,336         inc       rax                                           
00,337         mov       QWORD PTR [40+rsp], rax                       
00,338         cmp       rax, QWORD PTR [800064+rsp]                   
00,339         jg        .B4.39        
00,340 .B4.30::                        
00,341         mov       rbx, QWORD PTR [40+rsp]                       
00,342         test      rbx, rbx                                      
00,343         jle       .B4.28        
00,344 .B4.32::                        
00,345         mov       rax, QWORD PTR [-8+r13+rbx*8]                 
00,346         cmp       rax, QWORD PTR [r13+rbx*8]                    
00,347         jbe       .B4.28        
00,348 .B4.33::                        
00,349         lea       rdx, QWORD PTR [r13+rbx*8]                    
00,350         lea       rcx, QWORD PTR [-8+rdx]                       
00,351         call      swapUnconditional                             
00,352 .B4.34::                        
00,353         dec       rbx                                           
00,354         test      rbx, rbx                                      
00,355         jg        .B4.32        
00,356         jmp       .B4.28        
00,357 .B4.39::                        
00,358         add       rsp, 800152                                   
00,359         pop       rbp                                           
00,360         pop       r15                                           
00,361         pop       r14                                           
00,362         pop       r13                                           
00,363         pop       r12                                           
00,364         pop       rdi                                           
00,365         pop       rsi                                           
00,366         pop       rbx                                           
00,367         ret                                                     
00,368 .B4.40::
00,369 Quicksort_QB64_v7 ENDP
00,370 
00,371 swapUnconditional	PROC 
00,372 .B5.1::                         
00,373         mov       rax, QWORD PTR [rdx]                          
00,374         mov       r8, QWORD PTR [rcx]                           
00,375         mov       QWORD PTR [rcx], rax                          
00,376         mov       QWORD PTR [rdx], r8                           
00,377 .B5.4::                         
00,378         ret                                                     
00,379 .B5.2::
00,380 swapUnconditional ENDP
*/
